#
# Module for geometric computations and generic manipulation of canvas.
#
############################################################
# xcoord -
# ycoord -
# Functions for passing from polar to cartesian coordinates.
#
# Parameters:
# - center	coordinate of the center (x or y)
# - radius	distance of the point wrt the center
# - angle	angle of the segment wrt the horizontal
#
proc xcoord { center radius angle } {
    return [expr $center + int(floor($radius*cos($angle) ))]
}

proc ycoord { center radius angle } {
    return [expr $center + int(floor($radius*sin($angle) ))]
}

############################################################
# Clip -
# Procedure to clip a segment (x0,y0)->(x1,y1) at a distance radius of
# the goal. Returns a list of two points representing the new extremity.
#
# Parameters:
# - x0		x coord of the initial point
# - y0		y coord of the initial point
# - x1		x coord of the target point
# - y1		y coord of the target point
# - radius	distance of the target point at which to clip
#
proc clip { x0 y0 x1 y1 radius } {
    global DEBUG

    if $DEBUG {
	puts -nonewline "clipping ($x0,$y0)-($x1,$y1) at $radius "
    }

    set dx [expr $x1-$x0]
    set dy [expr $y1-$y0]

    set Distance [expr hypot($dx,$dy)]

    set x2 [expr $x1-int(floor( $dx * $radius / $Distance ))]
    set y2 [expr $y1-int(floor( $dy * $radius / $Distance ))]
    
    if $DEBUG {
	puts "...($x2,$y2)"
    }

    return [list $x2 $y2]
}

############################################################
# BreakMidPoint -
# Function used to transform a line (x0,y0)->(x1,y1) in pair of lines
# with same endpoints but with a new middle point situated at
# a distance "radius" of the initial endpoint in the orthogonal direction.
# If radius is >0 this counterclockwise, if <0, clockwise.
# Returns the new middle point as a list.
#
# Parameters:
# - x0		the x coordinate of the origin
# - y0		the y coordinate of the origin
# - x1		the x coordinate of the destination
# - y1		the y coordinate of the destination
# - radius 	distance of the mid point wrt the initial middle
#
proc BreakMidPoint { x0 y0 x1 y1 radius } {

    #compute shifted  middle and put proba
    set dirx [expr $x1-$x0]
    set diry [expr $y1-$y0]
    set norm [expr sqrt($dirx*$dirx+$diry*$diry)]
    set dirx [expr $dirx/double($norm)]
    set diry [expr $diry/double($norm)]
    set xmid [expr ($x0+$x1)/2 +($diry*$radius)]
    set ymid [expr ($y0+$y1)/2 -($dirx*$radius)]

    return [list $xmid $ymid]
}

######################################################################
# BreakAndClipLine -
# Function to transform a line (x0,y0)->(x1,y1) as described
# in BreakMidPoint, but also clipping both ends at a distance clipradius.
# Returns the resulting line as a list with 6 elements
#
# Parameters:
# - x0		the x coordinate of the origin
# - y0		the y coordinate of the origin
# - x1		the x coordinate of the destination
# - y1		the y coordinate of the destination
# - shiftradius the distance of the mid point wrt the initial middle
# - clipradius	the distance at which to clip endpoints.
#
proc BreakAndClipLine { x0 y0 x1 y1 shiftradius clipradius } {
    
    set mid [BreakMidPoint $x0 $y0 $x1 $y1 $shiftradius]
    set xmid [lindex $mid 0]
    set ymid [lindex $mid 1]

    set end1 [clip $xmid $ymid $x0 $y0 $clipradius]
    set end2 [clip $xmid $ymid $x1 $y1 $clipradius]

    set res [concat $end1 [list $xmid $ymid] $end2]

    return $res
}

######################################################################
# ClipLine -
# Function to transform a given line by clipping both ends at a distance 
# clipradius.
# Returns the resulting line.
#
# Parameters:
# - ptlist	the line to clip as a list of points
# - clipradius	the distance at which to clip endpoints.
#
proc ClipLine { ptlist clipradius } {
    
    set end [clip [lindex $ptlist 2] [lindex $ptlist 3] \
		 [lindex $ptlist 0] [lindex $ptlist 1] $clipradius]
    set res [lreplace $ptlist 0 1 [lindex $end 0] [lindex $end 1]]
    set nbpts [llength $res]
    set end [clip [lindex $res [expr $nbpts-4]] \
		 [lindex $res [expr $nbpts-3]] \
		 [lindex $res [expr $nbpts-2]] \
		 [lindex $res [expr $nbpts-1]] $clipradius]
    set res [lreplace $res [expr $nbpts-2] [expr $nbpts-1] \
		 [lindex $end 0] [lindex $end 1]]
    
    return $res
}

############################################################
# Shift_Line_Coords -
# A function to shift a list of points (presumably a line) by (dx,dy).
# Returns the new list.
#
# Parameters:
# - line	a line: a list of (x,y) points
# - dx		by how much to shift in the x direction
# - dy		by how much to shift in the y direction
#
proc Shift_Line_Coords { line dx dy } {

    set newline ""

    set numpts [expr [llength $line] / 2]
    for { set i 0 } { $i < $numpts } { incr i 1 } {
	lappend newline [expr [lindex $line [expr 2*$i]]+$dx]
	lappend newline [expr [lindex $line [expr 2*$i+1]]+$dy]
    }

    return $newline
}

############################################################
# Scale_Line_Coords -
# A function to scale a list of points (presumably a line) by some factor.
# Returns the new list.
#
# Parameters:
# - line	a line: a list of (x,y) points
# - factor	the factor by how much to multiply the coordinates
#
proc Scale_Line_Coords { line factor } {

    set newline ""

    foreach val $line {
	lappend newline [expr $val*$factor]
    }

    return $newline
}

######################################################################
# Resize_Graph -
# Modify the extension (scrollregion) of a canvas according to specified
# dimensions. If a dimension is smaller than the current, nothing is 
# changed. Otherwise, the canvas is expanded.
#
# Parameters:
# - win		the canvas where this happens
# - newx	new x dimansion
# - newy	new y dimension
#
proc Resize_Graph { win newx newy } {
    
    set curzone [lindex [$win configure -scrollregion] 4]
    set curx [lindex $curzone 2]
    set cury [lindex $curzone 3]
    if [expr { $newx > $curx } || { $newy > $cury } ] {
	set xx [expr $newx>$curx ? $newx : $curx]
	set yy [expr $newy>$cury ? $newy : $cury]
	$win configure -scrollregion [list 0 0 $xx $yy]
	pack $win
    }

}

######################################################################
# The two following procedures (Geometry_Browser and UpdateGeomParams)
# allow to greate a Browser for the geometry of a set of generic nodes.
# The geometry itself is retuned in a global array named GeomParams. 
# The "fields" of this array are:
# - nodenb 	the total number of nodes in the set
# - xdim	the number of columns (case of a grid for instance)
# - ydim	the number of rows
# - current	current type 
# - orient	orientation (for strings for instance)
# - object_distance	typical distance between nodes
# - xdistance	horizontal spacing
# - ydistance	vertical spacing
# - node_radius	the radius of the circle representing the nodes
######################################################################
# Geometry_Browser -
# Creates a frame with a menu allowing to choose a geometry.
# The frame is attached to an existing one passed as a parameter.
# The name is returned as result.
#
# Parameter:
# - win		the window (frame) into which to pack the browser
#
proc Geometry_Browser { { win . } } {
    global GeomParams

    set f [frame $win.geom]
    menubutton $f.choice -text "Geometry:" -width 10 -underline 0 \
	-menu $f.choice.menu -anchor nw

    array set GeomName { circle Circle cycle Cycle complete Complete \
			     star Star random Random \
			     line Line string String \
			     grid Grid bipartite Bipartite }

    set m [menu $f.choice.menu]
    foreach name $GeomParams(validnames) {
	$m add command -label $GeomName($name) \
	    -command "UpdateGeomParams $f $name $GeomName($name)" \
	    -underline 0
    }

    label $f.namelabel -width 8 -pady 3 -anchor w \
	-text $GeomName($GeomParams(current))
    pack $f.choice $f.namelabel -side left -anchor n

    # Create the default parameter area
    UpdateGeomParams $f $GeomParams(current) $GeomName($GeomParams(current))

    # Do not pack the frame: the caller will do it.
    # Return the frame name.
    return $f

}

######################################################################
# UpdateGeomParams -
# The procedure that handles the changes in the geometry menu.
# It creates and packs various frames inside the frame passed as
# a parameter, according to the current geometry.
#
# Parameters:
# - win		the frame containing the geometry browser
# - name	the name of the just chosen geometry
# - labelname	the string to put in the "namelabel" field.
#
proc UpdateGeomParams { win name labelname } {
    global GeomParams

    set GeomParams(current) $name
    $win.namelabel configure -text $labelname

    # In order to diminish flashing effects, the higher level frames
    # will not be deleted. The "size", "offset" and "action" zones
    # are always there anyway. The other are unpacked to free the space.
    foreach zone { size dist chord proba orient layout } {
	if ![winfo exists $win.$zone] { 
	    frame $win.$zone
	} else {
	    foreach child [winfo children $win.$zone] {
		destroy $child
	    }
	}
    }
#     foreach zone { chord proba orient } {
# 	pack forget $win.$zone

#     }

    switch $name {
	circle -
	complete {
	    # these three types have a circle of nodes as base.
	    label $win.size.label -text "Number of Nodes:" -width 15 \
		-anchor w
	    entry $win.size.field -textvar GeomParams(nodenb) \
		-width 3 -relief sunken
	    pack $win.size.label $win.size.field -side left

	    label $win.dist.label -text "Node distance:" -width 15 \
		-anchor w
	    entry $win.dist.field -textvar GeomParams(object_distance) \
		-width 3 -relief sunken
	    pack $win.dist.label $win.dist.field -side left

	}
	random {
	    label $win.size.label -text "Number of Nodes:" -width 15 \
		-anchor w
	    entry $win.size.field -textvar GeomParams(nodenb) \
		-width 3 -relief sunken
	    pack $win.size.label $win.size.field -side left

	    label $win.dist.label -text "Node distance:" -width 15 \
		-anchor w
	    entry $win.dist.field -textvar GeomParams(object_distance) \
		-width 3 -relief sunken
	    pack $win.dist.label $win.dist.field -side left

	    label $win.proba.label -text "Edge probability:" \
		-width 15 -anchor w
	    entry $win.proba.field -textvar GeomParams(proba) \
		-width 12 -relief sunken
	    pack $win.proba.label $win.proba.field -side left
	}
	cycle {
	    # this type is a circle, but with chords.
	    label $win.size.label -text "Number of Nodes:" -width 15 \
		-anchor w
	    entry $win.size.field -textvar GeomParams(nodenb) \
		-width 3 -relief sunken
	    pack $win.size.label $win.size.field -side left

	    label $win.dist.label -text "Node distance:" -width 15 \
		-anchor w
	    entry $win.dist.field -textvar GeomParams(object_distance) \
		-width 3 -relief sunken
	    pack $win.dist.label $win.dist.field -side left

	    label $win.chord.label -text "List of chords:" -width 15 -anchor w
	    entry $win.chord.field -textvar GeomParams(chordlist) \
		-width 10 -relief sunken
	    pack $win.chord.label -side left
	    pack $win.chord.field -side left -expand true -fill x
	}
	line -
	string {
	    set f $win.size
	    label $f.label -text "Number of Nodes:" -width 15 -anchor w
	    entry $f.field -textvar GeomParams(nodenb) \
		-width 3 -relief sunken
	    pack $f.label $f.field -side left

	    label $win.dist.label -text "Node distance:" -width 15 \
		-anchor w
	    entry $win.dist.field -textvar GeomParams(object_distance) \
		-width 3 -relief sunken
	    pack $win.dist.label $win.dist.field -side left

	    set f $win.orient
	    label $f.label -text "Orientation:" -width 12  -anchor w
	    radiobutton $f.hor -text "Horizontal" \
		-variable GeomParams(orient) -value hor
	    radiobutton $f.ver -text "Vertical"   \
		-variable GeomParams(orient) -value ver
	    radiobutton $f.dia -text "Diagonal"   \
		-variable GeomParams(orient) -value dia
	    pack $f.label $f.hor $f.ver $f.dia -side left -fill x

	}
	star {
	    label $win.size.label -text "Number of Branches:" -width 18 \
		-anchor w
	    entry $win.size.field -textvar GeomParams(nodenb) \
		-width 3 -relief sunken
	    pack $win.size.label $win.size.field -side left

	    label $win.dist.label -text "Node distance:" -width 15 \
		-anchor w
	    entry $win.dist.field -textvar GeomParams(object_distance) \
		-width 3 -relief sunken
	    pack $win.dist.label $win.dist.field -side left
	}
	grid { 
	    label $win.size.xlabel -text "Number of columns:" -width 18 \
		-anchor w
	    entry $win.size.xfield -textvar GeomParams(xdim) \
		-width 3 -relief sunken
	    label $win.size.ylabel -text "Number of rows:" -width 15 \
		-anchor w
	    entry $win.size.yfield -textvar GeomParams(ydim) \
		-width 3 -relief sunken
	    pack $win.size.xlabel $win.size.xfield \
		$win.size.ylabel $win.size.yfield \
		-side left

	    label $win.dist.xlabel -text "Distance of columns:" -width 18 \
		-anchor w
	    entry $win.dist.xfield -textvar GeomParams(xdistance) \
		-width 3 -relief sunken
	    label $win.dist.ylabel -text "Distance of rows:" -width 15 \
		-anchor w
	    entry $win.dist.yfield -textvar GeomParams(ydistance) \
		-width 3 -relief sunken
	    pack $win.dist.xlabel $win.dist.xfield \
		$win.dist.ylabel $win.dist.yfield \
		-side left
	}
	bipartite { 
	    label $win.size.left -text "Nodes at the left:" -width 18 \
		-anchor w
	    entry $win.size.lval -textvar GeomParams(leftnb) \
		-width 3 -relief sunken
	    label $win.size.right -text "Nodes at the right:" -width 18 \
		-anchor w
	    entry $win.size.rval -textvar GeomParams(rightnb) \
		-width 3 -relief sunken

	    label $win.dist.label -text "Node distance:" -width 15 \
		-anchor w
	    entry $win.dist.field -textvar GeomParams(object_distance) \
		-width 3 -relief sunken
	    pack $win.dist.label $win.dist.field -side left

	    set f $win.layout
	    label $f.label -text "Layout:" -width 12  -anchor w
	    radiobutton $f.hor -text "Horizontal" \
		-variable GeomParams(layout) -value hor
	    radiobutton $f.ver -text "Vertical"   \
		-variable GeomParams(layout) -value ver
	    radiobutton $f.cir -text "Circle"   \
		-variable GeomParams(layout) -value cir
	    pack $f.label $f.hor $f.ver $f.cir -side left -fill x

	    pack $win.size.left $win.size.lval \
		$win.size.right $win.size.rval \
		$win.dist.label $win.dist.field \
		-side left
	}
    }

    pack $win.size -side top -anchor w
    pack $win.dist -side top -anchor w
    if { $name == "cycle" } {
	pack $win.chord -side top -expand true -fill x -anchor w
    } else {
	pack forget $win.chord
    }
    if { $name == "random" } {
	pack $win.proba -side top -expand true -fill x -anchor w
    } else {
	pack forget $win.proba
    }
    if { $name == "line" } {
	pack $win.orient -side top -expand true -fill x -anchor w
    } else {
	pack forget $win.orient
    }
    if { $name == "bipartite" } {
	pack $win.layout -side top -expand true -fill x -anchor w
    } else {
	pack forget $win.layout
    }
	
}

