#
# Graph creation module for mced
#
# Two possibilities are given to the user:
# - create a set of nodes with some geometry but no arcs
# - load a chain from a file with a given geometry
#
######################################################################
# Make_From_Geom -
# Creates a user dialog interface for setting up the parameters before
# the creation of a set of nodes with geometry.
#
# Parameters:
# - canvas		the canvas window in which to create the objects
# 
proc Make_From_Geom { canvas } {
    global GeomParams

    set win .params

    if [catch {toplevel $win}] { return }

    wm title $win "Create Chain"

    Geometry_Browser $win

    # a separator for the looks
    frame $win.sep -relief sunken -height 2

    # a zone for specifying the offset
    frame $win.offset
    label $win.offset.label -text "Offset (mm)" -width 17 -anchor w
    label $win.offset.xlabel -text "X=" -width 3 -anchor w
    entry $win.offset.xoffset -textvar GeomParams(xoffset) \
	-width 4 -relief sunken
    label $win.offset.ylabel -text "Y=" -width 3 -anchor w
    entry $win.offset.yoffset -textvar GeomParams(yoffset) \
	-width 4 -relief sunken
    pack $win.offset.label $win.offset.xlabel $win.offset.xoffset \
	$win.offset.ylabel $win.offset.yoffset -side left 

    # the zone for buttons
    frame $win.actions
    button $win.actions.ok -text "New" \
	-underline 6 \
	-command "destroy [winfo toplevel $win]\n Create_From_Geom $canvas 0"
    button $win.actions.merge -text "Merge" \
	-underline 0 \
	-command "destroy [winfo toplevel $win]\n Create_From_Geom $canvas 1"
    button $win.actions.cancel -text "Cancel" \
	-underline 0 \
	-command "destroy [winfo toplevel $win]\n array set GeomParams {[array get GeomParams]}"
    pack $win.actions.ok $win.actions.merge $win.actions.cancel -side left

    # stacking the frames
    pack $win.geom    -side top -expand true -fill x
    pack $win.sep     -side top -expand true -fill x
    pack $win.offset  -side top -expand true -fill x
    pack $win.actions -side top -anchor center

    # standard bindings
    bind $win <Return> {$win.actions.ok invoke}
    bind $win <Escape> {$win.actions.cancel invoke}

}

######################################################################
# Create_From_Geom -
# The dipatching procedure for actual creation routines.
# It uses the values of GeomParams which have presumably been set
# with the geometry browser.
# Note that some parameters are passed on to the procedures, which
# use directly other parameters (offsets) from the global GeomParams!
# Returns a boolean value telling the success of the operation. Useful
# in Create_From_Matrix.
#
# Parameters:
# - win		the canvas window in which to create the objects
# - merge	a boolean telling whether the creation should be merged.
# 
proc Create_From_Geom { win merge } {
    global GeomParams

    if !$merge {
	if ![Clear_Graph $win] { return 0 }
    }

    switch $GeomParams(current) {
	circle {
	    create_circle $win $GeomParams(nodenb)
	}
	star {
	    create_star $win $GeomParams(nodenb)
	}
	line {
	    create_string $win $GeomParams(nodenb) $GeomParams(orient)
	}
	grid {
	    create_grid $win $GeomParams(xdim) $GeomParams(ydim)
	}
    }

    return 1
}

######################################################################
# - create_circle -
# - create_star -
# - create_string -
# - create_grid -
# Procedures for actually creating sets of nodes arranged according
# to some geometry.
# All these functions fill the "node_id" array with the ids of the nodes
# created, indexed from 0 to the total number minus one.
#
# Parameters
# - win		the canvas where the nodes are created
# - others	easy to see
#
proc create_circle { win size } {
    global GeomParams node_id PI 
    global DEBUG

    if { $size <= 0 } { return }

    set Alpha [expr 2*$PI / $size];
    if [ expr { sin($Alpha/2.0) == 0.0 } ] {
	set Radius 0.0
    } else {
	set Radius [expr $GeomParams(object_distance) / sin( $Alpha/2 )]
    }

    # putting the center... at the center
    set val $GeomParams(xoffset)
    set xcenter [expr [winfo pixels . ${val}m]+$GeomParams(object_distance)+$Radius]
    set val $GeomParams(yoffset)
    set ycenter [expr [winfo pixels . ${val}m]+$GeomParams(object_distance)+$Radius]

    # perhaps it is necessary to resize the canvas?
    Resize_Graph $win [expr $xcenter+$Radius+$GeomParams(object_distance)] \
	[expr $ycenter+$Radius+$GeomParams(object_distance)]

    if $DEBUG {
	puts "Creating a circle with $size nodes."
    }

    # creation of the nodes 
    for { set i 0 } { $i < $size } { incr i 1 } {
	set xpos [xcoord $xcenter $Radius [expr $i*$Alpha]]
	set ypos [ycoord $ycenter $Radius [expr $i*$Alpha]]

	set node_id($i) [make_new_node $win $xpos $ypos]
    }
}

proc create_star { win nb_branches } {
    global GeomParams node_id PI
    global DEBUG 

    # size will be the number of nodes on the circle
    set size [expr $nb_branches-1]

    if { $size <= 0 } { return }

    if $DEBUG {
	puts -nonewline "Creating of a star with $nb_branches branches."
    }

    # the creation of the circle will be handled by create_circle
    # but we have to anticipate where the center will be!
    set Alpha [expr 2*$PI / $size];
    if [ expr { sin($Alpha/2.0) == 0.0 } ] {
	set Radius 0.0
    } else {
	set Radius [expr $GeomParams(object_distance) / sin( $Alpha/2 )]
    }

    create_circle $win $size

    # determine location of the center
    set val $GeomParams(xoffset)
    set xorig [expr [winfo pixels . ${val}m] \
		   +$GeomParams(object_distance)+$Radius]
    set val $GeomParams(yoffset)
    set yorig [expr [winfo pixels . ${val}m] \
		   +$GeomParams(object_distance)+$Radius]

    # make center
    set node_id($size) [make_new_node $win $xorig $yorig]

}

proc create_string { win size orient } {
    global GeomParams node_id
    global DEBUG

    if $DEBUG {
	puts -nonewline "Creating of a line in direction $dir with $size nodes."
    }

    set distance $GeomParams(object_distance)

    # defining the origin of the bounding rectangle
    set val $GeomParams(xoffset)
    set xorig [expr [winfo pixels . ${val}m]+$distance]
    set val $GeomParams(yoffset)
    set yorig [expr [winfo pixels . ${val}m]+$distance]

    # Computing first the position of each node
    for { set i 0 } { $i < $size } { incr i 1 } {

        case $orient {
            {hor} {
                set X($i) [expr $xorig + $distance*$i]
                set Y($i) $yorig
	    }
            {ver} {
		set X($i) $xorig
		set Y($i) [expr $yorig + $distance*$i]
	    }
            {dia} {
		set X($i) [expr $xorig + $distance*$i]
		set Y($i) [expr $yorig + $distance*$i]
	    }
	}
    }
    # find out if resizing is necessary
    set xmax [TableMax X 0 [expr $size-1]]
    set ymax [TableMax Y 0 [expr $size-1]]
    Resize_Graph $win [expr $xmax+$distance] [expr $ymax+$distance]

    # now put the objects...
    for { set i 0 } { $i < $size } { incr i } {

	set node_id($i) [make_new_node $win $X($i) $Y($i)]
    }
}

proc create_grid { win xdim ydim } {
    global GeomParams node_id
    global DEBUG

    if $DEBUG {
	puts "Creating a $xdim x $ydim grid."
    }

    set object_distance $GeomParams(object_distance)

    # putting the origin in the NW corner
    set val $GeomParams(xoffset)
    set xorig [expr [winfo pixels . ${val}m]+$GeomParams(object_distance)]
    set val $GeomParams(yoffset)
    set yorig [expr [winfo pixels . ${val}m]+$GeomParams(object_distance)]

    # perhaps it is necessary to resize the canvas?
    Resize_Graph $win [expr $xorig+$xdim*$GeomParams(xdistance)] \
	[expr $yorig+$ydim*$GeomParams(ydistance)]

    # put nodes on the grid, row by row
    set num 0
    for { set i 0 } { $i < $xdim } { incr i 1 } {
	for { set j 0 } { $j < $ydim } { incr j 1 } {
	    set xpos($num) [expr $xorig+$i*$GeomParams(xdistance)]
	    set ypos($num) [expr $yorig+$j*$GeomParams(ydistance)]

	    set node_id($num) [make_new_node $win $xpos($num) $ypos($num)]
	    incr num
	}
    }
}

######################################################################
# Get_Matrix_From_File - 
# A function reading a matrix from a file and returning it:
# - either as a list of triples (i,j,Aij) (sparse matrix representation)
# - or as a full block
# The result is a 4-tuple (size,type,reward,matrix) where the type is one of
# "sparse" or "full", and reward is a boolean.
# The format of the input matrix is the MCL but the header has been eliminated
# until the size (included).
#
# Parameters:
# - file_name	the name of the file to read
# - sizeonly	boolean: if true, return just the size of the matrix
#
proc Get_Matrix_From_File { file_name { sizeonly 0 } } {
    global DEBUG

    if [catch {set matfile [open $file_name r]}] {
	ReportError "Create Error" "Can't open file $file_name."
	return 0
    }

    # read the FIRST LINE of the file to check that it is a MCL file
    # and find out its properties. Defaults are continuous, full, no reward
    set mattype "full"
    set reward 0
    
    set info [gets $matfile]
    foreach token [split $info] {
	switch $token {
	    discrete   { set timetype "discrete" }
	    continuous { set timetype "continuous" }
	    reward     { set reward 1 }
	    sparse     { set mattype "sparse" }
	    full       { set mattype "full" }
	}
    }
    if ![info exists timetype] {
	ReportError "Create Error" "The file $file_name does not contain a Markov Chain"
	return 0
    }

    if [expr { $mattype == "full" } && $reward ] {
	ReportError "Create Error" "Full matrices unsupported for reward markov chains"
    }

    if { $DEBUG } {
	puts "Reading a $mattype $timetype time Markov Chain from file $file_name."
    }


    set size [gets $matfile]

    if $sizeonly {
	return $size
    }

    set matrix [read $matfile]
    # could try to get rid of "stop", no?

    return [list $size $mattype $reward $matrix]
}

######################################################################
# Create_From_Matrix -
# The procedure for the actual creation of a graph from a matrix.
# It reads a matrix then arranges it according to a geometry already
# specified.
#
# Parameters:
# - win		the canvas window in which to create the objects.
# - file_name	the name of the file from which to read the matrix.
# - merge	a boolean telling whether the creation should be merged.
# 
proc Create_From_Matrix { win file_name { merge 0 } } {
    global node_id node_list arc_list arc_proba
    global GeomParams
    global DEBUG

    # Read the matrix. This will check whether the file is good.
    set result [Get_Matrix_From_File $file_name] 
    set matsize [lindex $result 0]
    set mattype [lindex $result 1]
    set SPARSE ![string compare $mattype "sparse"]
    set reward  [lindex $result 2]
    set matrix  [lindex $result 3]
    if { $matsize <= 0 } { return }

    # Checking if the matrix fits, then create the nodes, handling the
    # clear/merge issue in the process.
    # Upon success, this will have created the node_id array.
    if { $GeomParams(current) == "grid" } {
	set the_nb [expr $GeomParams(xdim) * $GeomParams(ydim)]
    } else {
	set the_nb $GeomParams(nodenb)
    }

    if { $the_nb >= $matsize } {
	if ![Create_From_Geom $win $merge] { return }
    } else {
	ReportError "Create Error" "Not enough nodes for this matrix.\nThe matrix has $matsize states."
	return
    }

    # Recovering the coordinates of the nodes from the global variables
    # Perhaps passing the arrays xpos any ypos, computed in the create_xxx
    # procedures, would save this but I prefer to avoid passing globals...
    foreach node $node_list {
	set id [lindex $node 0]
	set X($id) [lindex $node 1]
	set Y($id) [lindex $node 2]
    }
    for { set i 0 } { $i < $matsize } { incr i 1 } {
	set xpos($i) $X($node_id($i))
	set ypos($i) $Y($node_id($i))
    }

    if !$SPARSE {
	# draw arcs according to the matrix
	for { set i 0 } { $i < $matsize } { incr i 1 } {
	    set xpos1 $xpos($i)
	    set ypos1 $ypos($i)

	    for { set j 0 } { $j < $matsize } { incr j 1 } {
		# get value
		switch $type {
		    1 {
			set val [lindex $matrix [expr $i*$matsize+$j]]
		    }
		    2 {
			if [expr $i < $j] {
			    set val [lindex $matrix [expr $i*$matsize+$j-($i-1)]]
			} else {
			    set val [lindex $matrix [expr $i*$matsize+$j-$i]]
			}
		    }
		}
		if [expr $val != "0.0" ] {

		    set xpos2 $xpos($j)
		    set ypos2 $ypos($j)

		    Put_Arc $win $node_id($i) $node_id($j) $val 0.0 \
			$xpos1 $ypos1 $xpos2 $ypos2
		}
	    }
	}
    } else {
	# matrix is sparse here. Should take care of rewards if any.
	set idx 0

	set matlength [llength $matrix]
	for { set CONTINUE 1 } { $CONTINUE } {} {
	    if { $idx < $matlength } {
		set CONTINUE [expr 0 != \
				  [string compare [lindex $matrix $idx] \
				       "stop"]]
	    }
	    set CONTINUE [expr $CONTINUE && ($idx+2 < $matlength)]

	    if $CONTINUE {
		# puts "New row at idx $idx"
		set orig [lindex $matrix $idx]
		incr idx
		set dest [lindex $matrix $idx]
		incr idx
		set val [lindex $matrix $idx]
		incr idx
		
		# default reward is 0.0
		set rew 0.0

		if $reward {
		    # puts "Got ($orig,$dest). Looking for reward in [lrange $matrix $idx end]"
		    # reading a token which should be "wr"
		    set tok [lindex $matrix $idx]
		    incr idx
		    if { $tok == "wr" } {
			set rew [lindex $matrix $idx]
			incr idx
		    } else {
			set rew $tok
		    }
		}

		set xpos1 $xpos($orig)
		set ypos1 $ypos($orig)
		set xpos2 $xpos($dest)
		set ypos2 $ypos($dest)

		Put_Arc $win $node_id($orig) $node_id($dest) $val $rew \
		    $xpos1 $ypos1 $xpos2 $ypos2
	    }
	}
    }
    if $DEBUG {
	puts " ...Done"
    }
}

######################################################################
# Make_From_Matrix -
# Creates a user dialog interface for choosing a Markov Chain matrix 
# file and setting up the parameters before the creation of the representation.
# First, the name is chosen with a file browser, then only the geometry
# browser is proposed.
#
# Parameters:
# - canvas	the canvas window in which to create the objects
# - promptfile	a boolean controlling whether the user is prompted for a file
# - filename	a (proposed) file name for loading
# 
proc Make_From_Matrix { canvas { promptfile 0 } { file_name "" } } {
    global GeomParams

    set types {
	{{Markov Chain Language File}       {.mcl .mc}  }
	{{All Files}        *             }
    }

    if $promptfile {
	#     set file_name [tk_getOpenFile -defaultextension mcl \
	# 		       -filetypes $types -parent .]
	set options "-defaultextension mcl -filetypes {$types} -parent ."

	set file_name [GetOpenFile $options]
    }

    if { $file_name == "" } { return }

    # Read size first. This will check whether the file is good.
    set matsize [Get_Matrix_From_File $file_name 1]
    if { $matsize <= 0 } { return }

    # Now go for the geometry browser copied from Make_From_Geom
    if [catch {set win [toplevel .geometry]}] { return }
    wm title $win "Geometry of the Chain"

    set GeomParams(nodenb) $matsize
    Geometry_Browser $win

    # a separator for the looks
    frame $win.sep -relief sunken -height 2

    # a zone for specifying the offset
    frame $win.offset
    label $win.offset.label -text "Offset (mm)" -width 17 -anchor w
    label $win.offset.xlabel -text "X=" -width 3 -anchor w
    entry $win.offset.xoffset -textvar GeomParams(xoffset) \
	-width 4 -relief sunken
    label $win.offset.ylabel -text "Y=" -width 3 -anchor w
    entry $win.offset.yoffset -textvar GeomParams(yoffset) \
	-width 4 -relief sunken
    pack $win.offset.label $win.offset.xlabel $win.offset.xoffset \
	$win.offset.ylabel $win.offset.yoffset -side left 

    # the zone for buttons
    frame $win.actions
    button $win.actions.ok -text "New" \
	-underline 6 \
	-command "destroy [winfo toplevel $win]\n Create_From_Matrix $canvas $file_name 0"
    button $win.actions.merge -text "Merge" \
	-underline 0 \
	-command "destroy [winfo toplevel $win]\n Create_From_Matrix $canvas $file_name 1"
    button $win.actions.cancel -text "Cancel" \
	-underline 0 \
	-command "destroy [winfo toplevel $win]\n array set GeomParams {[array get GeomParams]}"
    pack $win.actions.ok $win.actions.merge $win.actions.cancel -side left

    # stacking the frames
    pack $win.geom    -side top -expand true -fill x
    pack $win.sep     -side top -expand true -fill x
    pack $win.offset  -side top -expand true -fill x
    pack $win.actions -side top -anchor center

    # standard bindings
    bind $win <Return> "$win.actions.ok invoke"
    bind $win <Escape> "$win.actions.cancel invoke"

}

######################################################################
# Put_Arc -
# Procedure for drawing an arc between two nodes and displaying its
# proba. When the nodes are different, makes a bent arc. When they are
# equal, makes a sort of loop.
#
# Parameters:
# - win		the canvas where the drawing takes place
# - orig_id	id of the origin node
# - dest_id	id of the destination node
# - val		proba/rate of the transition
# = reward	reward associated with the transition
# - x1		x coordinate of the origin
# - y1		y coordinate of the origin
# - x2		x coordinate of the destination
# - y3		y coordinate of the destination
#
proc Put_Arc { win orig_id dest_id val reward x1 y1 x2 y2 } {
    global GeomParams
    global arc_list arc_proba arc_reward

    set distance $GeomParams(object_distance)
    set node_radius $GeomParams(node_radius)
    # Drawing the arc is tricky. Depends on origin/destination.
    if { $orig_id != $dest_id } {

	set theline [BreakAndClipLine $x1 $y1 $x2 $y2 \
			 [expr $distance/4] $node_radius]

    } else {
	# drawing rudimentary self loop
	set end1 [clip [expr $x1+100] $y1 $x1 $y1 $node_radius]
	set end2 [clip $x1 [expr $y1-100] $x1 $y1 $node_radius]
	set theline [concat $end1 \
			 [expr $x1+$distance/5] $y1 \
			 [expr $x1+$distance/5] [expr $y1-$distance/5] \
			 $x1 [expr $y1-$distance/5] \
			 $end2]
    }

    set newid [eval {$win create line} $theline -tag ARC -arrow last]

    set new_arc [list $newid $orig_id $dest_id]
    lappend arc_list $new_arc
    set arc_proba($newid) $val
    set arc_reward($newid) $reward
	
    Display_Proba $win $newid
}

######################################################################
# Old_Way_Of_Creating_From_Matrix -
# OBSOLETE part of procedure: replaced by the current "Create_From_Matrix".
# Kept in case of bad surprise. 
#
# Parameters: none
#
proc Old_Way_Of_Creating_From_Matrix { win file_name { merge 0 } } {

    global GeomParams

    set object_distance $GeomParams(object_distance)

    switch $layout {
	circle {
	    set alpha [expr 2*$PI / $size];
	    if [ expr { sin($alpha/2.0) == 0.0 } ] {
		set radius 0.0
	    } else {
		set radius [expr $object_distance / 2 / sin( $alpha/2 )]
	    }

	    # putting the center... at the center
	    set val $GeomParams(xoffset)
	    set xorig [expr [winfo pixels . ${val}m]+$object_distance+$radius]
	    set val $GeomParams(yoffset)
	    set yorig [expr [winfo pixels . ${val}m]+$object_distance+$radius]

	    # perhaps it is necessary to resize the canvas?
	    Resize_Graph $win [expr $xorig+$radius+$object_distance] \
		[expr $yorig+$radius+$object_distance]

	    # put nodes on a circle
	    for { set i 0 } { $i < $size } { incr i 1 } {
		set xpos($i) [xcoord $xorig $radius [expr $i*$alpha]]
		set ypos($i) [ycoord $yorig $radius [expr $i*$alpha]]

		set no_id($i) [make_new_node $win $xpos($i) $ypos($i)]
	    }
	}
	grid {
	    set xdim [lindex $dimensions 0]
	    set ydim [lindex $dimensions 1]

	    # putting the origin in the NW corner
	    set val $GeomParams(xoffset)
	    set xorig [expr [winfo pixels . ${val}m]+$object_distance]
	    set val $GeomParams(yoffset)
	    set yorig [expr [winfo pixels . ${val}m]+$object_distance]

	    # perhaps it is necessary to resize the canvas?
	    Resize_Graph $win [expr $xorig+$xdim*$object_distance] \
		[expr $yorig+$ydim*$object_distance]

	    # put nodes on the grid, row by row
	    set num 0
	    for { set i 0 } { $i < $xdim } { incr i 1 } {
		for { set j 0 } { $j < $ydim } { incr j 1 } {
		    if { $num < $size } {
			set xpos($num) [expr $xorig+$i*$object_distance]
			set ypos($num) [expr $yorig+$j*$object_distance]

			set no_id($num) [make_new_node $win \
					     $xpos($num) $ypos($num)]
			incr num
		    }
		}
	    }
	}
	line {
	    # putting the origin in the NW corner
	    set val $GeomParams(xoffset)
	    set xorig [expr [winfo pixels . ${val}m]+$object_distance]
	    set val $GeomParams(yoffset)
	    set yorig [expr [winfo pixels . ${val}m]+$object_distance]

	    # perhaps it is necessary to resize the canvas?
	    Resize_Graph $win [expr $xorig+$size*$object_distance] \
		[expr $yorig+$object_distance]

	    # put nodes in a horizontal line
	    for { set i 0 } { $i < $size } { incr i 1 } {
		set xpos($i) [expr $xorig+$i*$object_distance]
		set ypos($i) $yorig

		set no_id($i) [make_new_node $win $xpos($i) $ypos($i)]
	    }
	}
    }

}