#
# Module grouping the procedures for "quantitative" analysis of 
# Markov chains 
#
# - exact numerical solution
# - simulation
# - graph simplification
#
######################################################################
# Run_Simulation -
# Interface to the call of "mc_sim" which simulates the Markov Chain.
#
# Parameters: 
# - sim_length	the number of jumps to simulate.
#
proc Run_Simulation { SimParams } {
    upvar $SimParams Parameters 
    global KEEP

    set TMP_MC_FILE "/tmp/mced[pid].mc"
    set TMP_RES_FILE "/tmp/mced[pid].res"

    if [write_matrix $TMP_MC_FILE sparse] {
	# add initial state
	set mfile [open $TMP_MC_FILE a]
	puts $mfile "0"
	close $mfile

	set command "mc_sim -f $TMP_MC_FILE -tmax $Parameters(sim_length)"
	if { $Parameters(trajectory) == "no" } {
	    set command "$command -no_traj"
	}
	if { $Parameters(statistics) == "yes" } {
	    set command "$command -stats"
	}
	if [catch [eval exec "$command" > $TMP_RES_FILE] problem] {
	    ReportError "Exec Error" "Error executing $command:\n$problem"
	    return
	}
	
	set resfile [open $TMP_RES_FILE r]
	set result [read $resfile]
	close $resfile

	LogResult "Simulation Result" "$result"
    }

    if !$KEEP {
	RemoveFile $TMP_MC_FILE
	RemoveFile $TMP_RES_FILE
    }
}

######################################################################
# Simulate -
# Dialog interface to let the user specify parameters for the 
# simulation of the Markov Chain
#
# Parameters: none
#
proc Simulate { } {
    global SimParams

    if [catch {toplevel .params}] { return }

    wm title .params "Simulation Parameters"

    frame .params.simlength
    label .params.simlength.label -text "Simulation Length" -width 15 -anchor w
    entry .params.simlength.field -textvar SimParams(sim_length) \
	-width 8 -relief sunken
    pack .params.simlength.label .params.simlength.field -side left \
	-expand true -fill x
    
    set f [frame .params.traj]
    label $f.label -text "Trajectory" -width 10 -anchor w 
    radiobutton $f.yes -text "Yes" -variable SimParams(trajectory) \
	-value yes
    radiobutton $f.no -text "No" -variable SimParams(trajectory) \
	-value no
    pack $f.label $f.yes $f.no -side left
    set f [frame .params.stats]
    label $f.label -text "Statistics" -width 10 -anchor w 
    radiobutton $f.yes -text "Yes" -variable SimParams(statistics) \
	-value yes
    radiobutton $f.no -text "No" -variable SimParams(statistics) \
	-value no
    pack $f.label $f.yes $f.no -side left

    frame .params.actions
    button .params.actions.ok -text "Ok" \
	-command "destroy .params\n Run_Simulation SimParams"
    button .params.actions.cancel -text "Cancel" \
	-command "destroy .params"
    
    pack .params.simlength -side top -fill x
    pack .params.traj -side top
    pack .params.stats -side top
    pack .params.actions.ok .params.actions.cancel -side left
    pack .params.actions -side top

}
    

proc run_solution { } {
    global KEEP

    set TMP_MC_FILE "/tmp/mced[pid].mc"
    set TMP_RES_FILE "/tmp/mced[pid].res"

    if [write_matrix $TMP_MC_FILE sparse] {
	# add initial state
	set mfile [open $TMP_MC_FILE a]
	puts $mfile "0"
	close $mfile

	exec mc_solve -f $TMP_MC_FILE  > $TMP_RES_FILE
	
	set resfile [open $TMP_RES_FILE r]
	set result [read $resfile]
	close $resfile

	LogResult "Markov Chain Analysis Result" "$result"
    }

    if !$KEEP {
	RemoveFile $TMP_MC_FILE
	RemoveFile $TMP_RES_FILE
    }
}

proc solve { } {

    # no interface necessary at this stage.
    run_solution
}

######################################################################
# Eliminate_Transients -
# A terse (semi-quantitative) procedure to eliminate states which are
# obviously transient from the graph.
# An obviously transient state is one which does not have an input,
# or just itself.
# The operation is repeated until nothing changes.
# This method does not allow to recognize transient classes.
# Note: for a better efficiency with very large graphs, the original
# incremental algorithm (node by node, deleting on the fly) has been
# replaced by a "batch" algorithm (identify removable nodes, then remove).
#
# Parameters:
# - win		the canvas in which this happens.
#
proc Eliminate_Transients { win } {
    global node_list arc_list

    set removed 1
    for { } { $removed > 0 } { } {
	set removed 0
	if [info exists has_input] {
	    unset has_input
	}
	foreach arc $arc_list {
	    set dest [lindex $arc 2]
	    if { $dest != [lindex $arc 1] } {
		set has_input($dest) 1
	    }
	}
	foreach node $node_list {
	    set nodeid [lindex $node 0]
	    if ![info exists has_input($nodeid)] {
		Delete_Object $win $nodeid
		incr removed 1
		update idletasks
	    }
	}
	puts "Removed $removed states"
    }
}

proc Old_Eliminate_Transients { win } {
    global node_list arc_list

    set removed 1
    for { } { $removed > 0 } { } {
	set removed 0
	# list of nodes is copied at the top of the loop because
	# it may change inside.
	set nodes $node_list
	foreach node $nodes {
	    set nodeid [lindex $node 0]
	    set idegree 0
	    foreach arc $arc_list {
		if [expr { $nodeid == [lindex $arc 2] } \
			&& { $nodeid != [lindex $arc 1] }] {
		    incr idegree 1
		}
	    }

	    if { $idegree == 0 } {
		Delete_Object $win $nodeid
		incr removed 1
		update idletasks
	    }
	}
	puts "Removed $removed states"
    }
}
