#
# Module containing procedures related to moving objects on the canvas.
#
########################################################################
# Move_Object -
# A procedure to move an object. 
#
# When called with mode=neighborhood (which is the default), it (supposedly) 
# takes care of moving related objects, such as:
# - marks in places
# - text
# - extremities of links.
# In this mode, links cannot be moved.
# When called with mode=alone, only the object itself is moved. 
#
# Parameters:
# - win		the canvas widget where this is happening
# - obj_id	the id of the object to move
# - dx		increment in the x direction (in pixels)
# - dy		increment in the y direction (in pixels)
# - mode	either "neighborhood" or "alone"
# - taboo	a list of objects NOT to move in conjunction with the
#               neighborhood mode.
#
proc Move_Object { win obj_id dx dy { mode neighborhood } { taboo {} } } {
    global trans_list place_list source_list
    global mark curzoomvalue place_radius
    global DEBUG

    if { "" == $obj_id } { return }

    if { $mode == "alone" } {
	$win move $obj_id $dx $dy 
	return
    }

    set objtags [$win gettags $obj_id]
    if $DEBUG {
	puts "Move: object $obj_id has tags $objtags"
    }

    set scale [expr $curzoomvalue/100.0]
    set pl_rad [expr $place_radius*$scale]

    # identify attached text objects
    set textlist [concat [$win find withtag "Tx$obj_id"] \
		      [$win find withtag "Lw$obj_id"]]

    if [Is_Type $obj_id PLACE] {
	# got to move all marks... so record them before moving the object
	set marklist [Get_Marks_Of_Place $win $obj_id]

	# get coords of place, for recomputing coords of center
	set goal [$win coords $obj_id]

	# find in/out links
	set i_links [Get_Place_Input_Links $obj_id]
	foreach link $i_links {
	    if { [lsearch $taboo $link] == -1 } {
		set pts [$win coords $link]
		set idx [expr [llength $pts]-1]
		set jdx [expr $idx-1]
		set segment [lrange $pts [expr $idx-3] $idx]
		set segment [lreplace $segment 2 3 \
				 [expr [lindex $goal 0]+$pl_rad+$dx] \
				 [expr [lindex $goal 1]+$pl_rad+$dy]]
		set end [eval {clip} $segment $pl_rad]
		# puts "Adj link $link: $segment EE $end"
		set pts [lreplace $pts $jdx $idx \
			     [lindex $end 0] [lindex $end 1]]
		eval {$win coords $link} $pts
		Display_Weight $win $link
	    }
	}

	set o_links [Get_Place_Output_Links $obj_id]
	foreach link $o_links {
	    if { [lsearch $taboo $link] == -1 } {
		set pts [$win coords $link]
		set segment [lrange $pts 0 3]
		set segment [list [lindex $segment 2] [lindex $segment 3] \
				 [expr [lindex $goal 0]+$pl_rad+$dx] \
				 [expr [lindex $goal 1]+$pl_rad+$dy]]
		set end [eval {clip} $segment $pl_rad]
		# puts "Adj link $link: $segment EE $end"
		set pts [lreplace $pts 0 1 \
			     [lindex $end 0] [lindex $end 1]]
		eval {$win coords $link} $pts
		Display_Weight $win $link
	    }
	}
    } elseif { [Is_Type $obj_id TRANS] || [Is_Type $obj_id SOURCE] \
		   || [Is_Type $obj_id SOURCE] } {
	# find in/out links
	set i_links [Get_Trans_Input_Links $obj_id]
	foreach link $i_links {
	    if { [lsearch $taboo $link] == -1 } {
		set pts [$win coords $link]
		set idx [expr [llength $pts]-1]
		set jdx [expr $idx-1]
		set pts [lreplace $pts $jdx $idx \
			     [expr [lindex $pts $jdx]+$dx] \
			     [expr [lindex $pts $idx]+$dy]]
		# puts "Adj link $link"
		eval {$win coords $link} $pts
		Display_Weight $win $link
	    }
	}

	set o_links [Get_Trans_Output_Links $obj_id]
	foreach link $o_links {
	    if { [lsearch $taboo $link] == -1 } {
		set pts [$win coords $link]
		set pts [lreplace $pts 0 1 \
			     [expr [lindex $pts 0]+$dx] \
			     [expr [lindex $pts 1]+$dy]]
		# puts "Adj link $link"
		eval {$win coords $link} $pts
		Display_Weight $win $link
	    }
	}
    }

    # move!
    $win move $obj_id $dx $dy 

    # Now only move tokens so that they appear on top.
    if [info exists marklist] {
	foreach token $marklist {
	    if [Is_Type $token MARK] {
		Move_Object $win $token $dx $dy $mode {}
	    }
	}
    }
    # finally move text
    if [info exists textlist] {
	foreach text $textlist {
	    if [Is_Type $text TEXT] {
		Move_Object $win $text $dx $dy $mode {}
	    }
	}
    }

    
}

proc Pick_Object_Or_Zone { win } {
    global cur_picked_obj
    global DEBUG

    set cur_obj [$win find withtag current]
    if $DEBUG {
	puts -nonewline "Pick: Current object = $cur_obj"
    }

    if { [Is_Type $cur_obj PLACE] || [Is_Type $cur_obj TRANS] \
	     || [Is_Type $cur_obj SOURCE] \
	     || [Is_Type $cur_obj SELZONE] } {
	if [info exists cur_picked_obj] {
	    # restore the color of currently picked object
	    $win itemconfigure $cur_picked_obj -outline black
	}
	# highlight new picked one
	$win itemconfigure $cur_obj -outline pink
	set cur_picked_obj $cur_obj
    }
}

proc Move_Object_Or_Zone { win dx dy } {
    global cur_picked_obj
    global DEBUG

    if ![info exists cur_picked_obj] { return }

    if $DEBUG {
	puts -nonewline "Shift of $dx $dy"
    }

    if { [Is_Type $cur_picked_obj PLACE] \
	     || [Is_Type $cur_picked_obj TRANS] \
	     || [Is_Type $cur_picked_obj SOURCE] } {
	Move_Object $win $cur_picked_obj $dx $dy neighborhood {}
    } elseif { [Is_Type $cur_picked_obj SELZONE] } {
	set bbox [$win bbox $cur_picked_obj]
	set targets [eval {$win find enclosed} $bbox]
	# extracting "taboo" links. These links will be moved by themselves
	# and will be "forgotten" when moving the objects at the end.
	set taboo ""
	foreach obj $targets {
	    if [Is_Type $obj LINK] {
		set props [Get_Link_Props $obj]
		set from [lindex $props 1]
		set to [lindex $props 2]
		# puts "?? Arc from $from to $to"
		if [expr { [lsearch $targets $from] >= 0 } \
			&& { [lsearch $targets $to] >= 0 } ] {
		    lappend taboo $obj
		}
	    }
	}
	
	if $DEBUG {
	    puts "Moving objects $targets (taboo list $taboo)"
	}
	foreach obj $targets {
	    if { [Is_Type $obj PLACE] || [Is_Type $obj TRANS] \
		     || [Is_Type $obj SOURCE] } {
		Move_Object $win $obj $dx $dy neighborhood $taboo
	    } 
	}
	foreach obj $taboo {
	    Move_Object $win $obj $dx $dy neighborhood $taboo
	}
	# moving zone itself
	$win move $cur_picked_obj $dx $dy

    } else {
	puts "Error: current object $cur_picked_obj should not be there."
    }
}

proc Drop_Object_Or_Zone { win xpos ypos } {
    global cur_picked_obj
    global DEBUG

    if ![info exists cur_picked_obj] { return }

    if $DEBUG {
	puts -nonewline "Drop at $xpos $ypos"
    }

    set bbox [$win bbox $cur_picked_obj]

    if { [Is_Type $cur_picked_obj PLACE] \
	     || [Is_Type $cur_picked_obj TRANS] \
	     || [Is_Type $cur_picked_obj SOURCE] } {
	set xfrom [lindex $bbox 0]
	set yfrom [lindex $bbox 1]
	set dx [expr $xpos-$xfrom]
	set dy [expr $ypos-$yfrom]
	Move_Object $win $cur_picked_obj $dx $dy neighborhood {}
    } elseif { [Is_Type $cur_picked_obj SELZONE] } {
	set xfrom [lindex $bbox 0]
	set yfrom [lindex $bbox 1]
	set dx [expr $xpos-$xfrom]
	set dy [expr $ypos-$yfrom]

	set targets [eval {$win find enclosed} $bbox]
	# extracting "taboo" links. These links will be moved by themselves
	# and will be "forgotten" when moving the objects at the end.
	set taboo ""
	foreach obj $targets {
	    if [Is_Type $obj LINK] {
		set props [Get_Link_Props $obj]
		set from [lindex $props 1]
		set to [lindex $props 2]
		# puts "?? Arc from $from to $to"
		if [expr { [lsearch $targets $from] >= 0 } \
			&& { [lsearch $targets $to] >= 0 } ] {
		    lappend taboo $obj
		}
	    }
	}
	
	if $DEBUG {
	    puts "Moving objects $targets (taboo list $taboo)"
	}
	foreach obj $targets {
	    if { [Is_Type $obj PLACE] || [Is_Type $obj TRANS] \
		     || [Is_Type $obj SOURCE] } {
		Move_Object $win $obj $dx $dy neighborhood $taboo
	    }
	}
	foreach obj $taboo {
	    Move_Object $win $obj $dx $dy neighborhood $taboo
	}
	# moving zone itself
	$win move $cur_picked_obj $dx $dy

    } else {
	puts "Error: current object $cur_picked_obj should not be there."
    }
}

