#
# Module for creation and manipulation of basic objects of the editor.
# 
######################################################################
# Make_Place -
# A procedure to create a place with specified attributes.
# Options recognized are
# -marking	optional, default 0
# -name		optional, default ""
# -lawparams	optional, default to a constant distribution with value 0
# -x		mandatory, x position
# -y		mandatory, y position
#
# Parameters:
# - win		the graph canvas where the place is to be created
# - options	the list of options
#
proc Make_Place { win options } {
    global place_radius curzoomvalue place_list mark PlaceTime name
    global state
    global DEBUG

    if $DEBUG {
	puts "Making a place with options $options"
    }

    set lawparams [list name constant value 0.0]
    set themarking 0
    set thename ""
    set scale [expr $curzoomvalue/100.0]

    for { set i 0 } { $i < [llength $options] } { incr i 1 } {
	set opt [lindex $options $i]
	switch -- $opt {
	    -marking { 
		incr i 1
		set themarking [lindex $options $i]
	    }
	    -lawparams { 
		incr i 1
		set lawparams [lindex $options $i]
	    }
	    -name { 
		incr i 1
		set thename [lindex $options $i]
	    }
	    -x {
		incr i 1
		set x [lindex $options $i]
	    }
	    -y {
		incr i 1
		set y [lindex $options $i]
	    }
	}
    }

    if ![expr { [info exists x] && [info exists y] }] {
	error "Create Error" "Could not create place: missing argument -x or -y."
    } else {
	# seems ok. Let's create.
	set newid [$win create oval \
		       [expr $x-$place_radius*$scale] \
		       [expr $y-$place_radius*$scale] \
		       [expr $x+$place_radius*$scale] \
		       [expr $y+$place_radius*$scale] \
		       -fill green -tag PLACE]

	lappend place_list [list $newid $x $y]
    
	set mark($newid)      $themarking
	set PlaceTime($newid) $lawparams
	set name($newid)      $thename

	Draw_Marking $win $newid
	Display_Name $win $newid
	Display_Timing $win $newid
	
	set state($win) modified

	return $newid
    }
}

######################################################################
# Make_Trans -
# A procedure to create a place with specified attributes.
# Options recognized are
# -name		optional, default ""
# -lawparams	mandatory, the firing time distribution
# -x		mandatory, x position
# -y		mandatory, y position
#
# Parameters:
# - win		the graph canvas where the place is to be created
# - options	the list of options
#
proc Make_Trans { win options } {
    global trans_width trans_height curzoomvalue trans_list
    global TransTime name
    global state
    global DEBUG

    if $DEBUG {
	puts "Making a transition with options $options"
    }

    set thename ""
    set scale [expr $curzoomvalue/100.0]

    for { set i 0 } { $i < [llength $options] } { incr i 1 } {
	set opt [lindex $options $i]
	switch -- $opt {
	    -lawparams {
		incr i 1
		set lawparams [lindex $options $i]
	    }
	    -name { 
		incr i 1
		set thename [lindex $options $i]
	    }
	    -x {
		incr i 1
		set x [lindex $options $i]
	    }
	    -y {
		incr i 1
		set y [lindex $options $i]
	    }
	}
    }

    if ![expr { [info exists x] && [info exists y] && [info exists lawparams] }] {
	error "Create Error" "Could not create transition: missing argument -x, -y or -lawparams"
    } else {
	# seems ok. Let's create.
	set newid [$win create rectangle \
		       [expr $x-$trans_width*$scale] \
		       [expr $y-$trans_height*$scale] \
		       [expr $x+$trans_width*$scale] \
		       [expr $y+$trans_height*$scale] \
		       -fill red -tag TRANS]

	lappend trans_list [list $newid $x $y]
    
	set TransTime($newid) 	$lawparams
	set name($newid) 	$thename

	Display_Name $win $newid
	Display_Timing $win $newid

	set state($win) modified

	return $newid
    }
}

######################################################################
# Make_Mark -
# A procedure to create a mark with specified attributes.
# Options recognized are
# -place	mandatory, the id of the place to which the mark belongs
# -x		mandatory, x position
# -y		mandatory, y position
#
# Parameters:
# - win		the graph canvas where the place is to be created
# - options	the list of options
#
proc Make_Mark { win options } {
    global curzoomvalue mark_radius

    set scale [expr $curzoomvalue/100.0]

    for { set i 0 } { $i < [llength $options] } { incr i 1 } {
	set opt [lindex $options $i]
	switch -- $opt {
	    -place { 
		incr i 1
		set theplace [lindex $options $i]
	    }
	    -x {
		incr i 1
		set x [lindex $options $i]
	    }
	    -y {
		incr i 1
		set y [lindex $options $i]
	    }
	}
    }

    if ![expr { [info exists x] && [info exists y] && [info exists theplace] }] {
	error "Create Error" \
	    "Could not create mark: missing argument -x or -y or -place."
    } else {
	# seems ok. Let's create.
	set tlist [list MARK "Pl$theplace"]
	set newid [$win create oval \
		       [expr $x-$mark_radius*$scale] \
		       [expr $y-$mark_radius*$scale] \
		       [expr $x+$mark_radius*$scale] \
		       [expr $y+$mark_radius*$scale] \
		       -fill black -tags $tlist ]

	set state($win) modified
	
	return $newid
    }
}

######################################################################
# Put_Place -
# Procedure to put a new place at the specified position.
# The attributes of the new object are the default attributes.
# Used in the "place mode".
#
# Parameters:
# - win		the canvas where to put the object
# - x		x position
# - y 		y position
#
proc Put_Place { win x y } {
    global default_props

    set new_id [Make_Place $win \
		    [list \
			 -x $x \
			 -y $y \
			 -marking $default_props(place_marking) \
			 -lawparams $default_props(PlaceTime) \
			 -name $default_props(place_name) \
			] ]

    if { $new_id == "" } {
	error "Create Error" "Could not create place at ($x,$y)"
    } else {
	return $new_id
    }
}

######################################################################
# Put_Trans -
# Procedure to put a new transition at the specified position.
# The attributes of the new object are the default attributes.
# Used in the "transition mode".
#
# Parameters:
# - win		the canvas where to put the object
# - x		x position
# - y 		y position
#
proc Put_Trans { win x y } {
    global default_props 

    set new_id [Make_Trans $win \
		    [list \
			 -x $x \
			 -y $y \
			 -lawparams $default_props(TransTime) \
			 -name      $default_props(place_name) \
			] ]

    if { $new_id == "" } {
	error "Create Error" "Could not create transition at ($x,$y)"
    } else {
	return $new_id
    }
}

######################################################################
# Put_Mark -
# Procedure to put a new mark at the specified position.
# Normally used indirectly since there is no "mark mode".
#
# Parameters:
# - win		the canvas where to put the object
# - place	the id of the place to which the mark is attached
# - x		x position
# - y 		y position
#
proc Put_Mark { win place x y } {

    set new_id [Make_Mark $win \
		    [list \
			 -place $place \
			 -x $x \
			 -y $y \
			] ]

    if { $new_id == "" } {
	error "Create Error" "Could not create mark for place $place at ($x,$y)"
    } else {
	return $new_id
    }
}

######################################################################
# Make_Source -
# A procedure to create a source with specified attributes.
# Options recognized are
# -name		optional, default ""
# -time		optional, default 0
# -x		mandatory, x position
# -y		mandatory, y position
#
# Parameters:
# - win		the graph canvas where the place is to be created
# - options	the list of options
#
proc Make_Source { win options } {
    global default_props source_radius source_list source_rate name
    global state

    if $DEBUG {
	puts "Making a source with options $options"
    }

    set thename ""
    set thetime 0
    for { set i 0 } { $i < [llength $options] } { incr i 1 } {
	set opt [lindex $options $i]
	switch -- $opt {
	    -time { 
		incr i 1
		set thetime [lindex $options $i]
	    }
	    -name { 
		incr i 1
		set thename [lindex $options $i]
	    }
	    -x {
		incr i 1
		set x [lindex $options $i]
	    }
	    -y {
		incr i 1
		set y [lindex $options $i]
	    }
	}
    }

    if ![expr { [info exists x] && [info exists y] }] {
	error "Create Error" "Could not create source: missing argument -x or -y"
    } else {
	# seems ok. Let's create.
	set newid [$win create rectangle \
		       [expr $x-$source_radius] [expr $y-$source_radius] \
		       [expr $x+$source_radius] [expr $y+$source_radius] \
		       -fill skyblue -tag SOURCE]

	lappend source_list [list $newid $x $y 0]

	set source_rate($newid) $thetime
	set name($newid)        $thename

	Display_Name $win $newid
	Display_Timing $win $newid
	
	set state($win) modified

	return $newid
    }
}

######################################################################
# Put_Source -
# Procedure to put a new mark at the specified position.
# Normally used indirectly since there is no "mark mode".
#
# Parameters:
# - win		the canvas where to put the object
# - place	the id of the place to which the mark is attached
# - x		x position
# - y 		y position
#
proc Put_Source { win x y } {
    global default_props source_radius source_list source_rate name
    global state

    set newid [Make_Source $win \
		   [list \
			-x $x \
			-y $y \
			-time $default_props(source_time) \
		       ] ]

    if { $new_id == "" } {
	error "Create Error" "Could not create source at ($x,$y)"
    } else {
	return $new_id
    }
}

#######################################################################
# Delete_Object -
# Removes an object from the canvas. Removes also the "attached" objects
# such as text, links.
# Handles the redirection of the attention to the "owner" of the object
# in case the object is a text (apparently not yet!).
#
# Parameters:
# - win		the canvas where this is happening
# - obj_id	the id of the object to remove with its "environment"
#
proc Delete_Object { win obj_id } {
    global trans_list place_list source_list platra_list trapla_list
    global mark
    global DEBUG

    if { "" == $obj_id } { return }

    set objtags  [$win gettags $obj_id]

    if $DEBUG {
	puts "DEL: object = $obj_id, tags = $objtags"
    }

    if [Is_Type $obj_id PLACE] {
	set place_list [Remove_From_List 0 $place_list $obj_id]
	Delete_From_List $win 1 $platra_list $obj_id
	Delete_From_List $win 2 $trapla_list $obj_id
	set platra_list [Remove_From_List 1 $platra_list $obj_id]
	set trapla_list [Remove_From_List 2 $trapla_list $obj_id]
	# got to remove all marks...
	set marklist [Get_Marks_Of_Place $win $obj_id]
	eval {$win delete} $marklist
    } elseif [Is_Type $obj_id TRANS] {
	set trans_list [Remove_From_List 0 $trans_list $obj_id]
	Delete_From_List $win 1 $trapla_list $obj_id
	Delete_From_List $win 2 $platra_list $obj_id
	set trapla_list [Remove_From_List 1 $trapla_list $obj_id]
	set platra_list [Remove_From_List 2 $platra_list $obj_id]
    } elseif [Is_Type $obj_id SOURCE] {
	set source_list [Remove_From_List 0 $source_list $obj_id]
	Delete_From_List $win 1 $trapla_list $obj_id
	set trapla_list [Remove_From_List 1 $trapla_list $obj_id]
    } elseif [Is_Type $obj_id LINK] {
	# don't know if it is platra or trapla: removing blindly...
	set platra_list [Remove_From_List 0 $platra_list $obj_id]
	set trapla_list [Remove_From_List 0 $trapla_list $obj_id]	    
    } elseif [Is_Type $obj_id MARK] {
	# find place to whom this belongs to decrease its marking.
	set place_id [Get_Place_Of_Mark $win $obj_id]
	if { $place_id != "" } {
	    if { $mark($place_id) >= 1 } {
		incr mark($place_id) -1
		Draw_Marking $win $place_id
	    } else {
		tkerror "Error: marking of place $place_id is already 0!"
	    }
	}
    }
 
    # if mark, already gone... 
    # if selection, should be destroyed with its object
    if ![expr [Is_Type $obj_id MARK] || [Is_Type $obj_id SELECT]] {
	$win delete $obj_id

	# delete text associated
	set objlist [$win find withtag "Tx$obj_id"]
	$win delete $objlist

	# delete weight associated
	set objlist [$win find withtag "Lw$obj_id"]
	$win delete $objlist

	# delete "selection" if any
	set objlist [$win find withtag "Sel$obj_id"]
	$win delete $objlist
    }

}

#######################################################################
# Delete_Current_Object -
# Removes the "current" object, usually the one pointed at. 
#
# Parameters:
# - win		the canvas where this is happening
#
proc Delete_Current_Object { win } {
    global DEBUG 

    if $DEBUG {
	puts "Current object = $cur_obj"
    }

    set cur_obj [$win find withtag current]
    if { [llength $cur_obj] > 1 } { 
	puts "Warning! More than one current object!"
    }
    if { [llength $cur_obj] > 0 } { 
	Delete_Object $win [lindex $cur_obj 0]
    }
}
