#
# Module for the editing of the objects: graphical modifications 
# and modification of attributes.
# 
######################################################################
# Edit_DepLink_Attributes -
# Create a user dialog frame for specifying the attributes of a 
# simple dependency link.
# The attributes are: its duration distribution.
#
# Parameters:
# - win		the canvas containing the object
# - link_id	the id of the object
# - Xpos	the x coordinate of the object, used to position the popup
# - Ypos	the y coordinate of the object
#
proc Edit_DepLink_Attributes { win link_id Xpos Ypos } {
    global LawParams LinkTime curzoomvalue object_distance

    if [catch {toplevel .att}] { return }

    wm title .att "Attributes of dependency link $link_id"

    wm geometry .att \
	"+[expr $Xpos+$object_distance*$curzoomvalue/100/2]+$Ypos"

    frame .att.time
    label .att.time.label -text "Duration" -width 10 -anchor w
    
    # create law browsing area, setting the parameters to that of 
    # the link
    array set LawParams $LinkTime($link_id)
    set lb [Law_Browser .att.time LawParams]

    frame .att.actions
    button .att.actions.ok -text "Set" \
	-command "destroy .att\n SetLinkLaw $link_id\n Display_Timing $win $link_id"
    button .att.actions.cancel -text "Cancel" \
	-command "destroy .att"
    
    pack .att.time.label -side left -pady 3 -anchor n
    pack $lb -side left
    pack .att.actions.ok .att.actions.cancel -side left -fill x
    pack .att.time .att.actions -side top

    bind .att <Return> {.att.actions.ok invoke}
    bind .att <Escape> {.att.actions.cancel invoke}

#    grab .att

}

######################################################################
# Edit_ProgLink_Attributes -
# Create a user dialog frame for specifying the attributes of a 
# program dependency link.
# The attributes are: its depth.
#
# Parameters:
# - win		the canvas containing the object
# - link_id	the id of the object
# - Xpos	the x coordinate of the object, used to position the popup
# - Ypos	the y coordinate of the object
#
proc Edit_ProgLink_Attributes { win link_id Xpos Ypos } {
    global link_lag name curzoomvalue object_distance

    if [catch {toplevel .att}] { return }

    wm title .att "Attributes of program link $link_id"

    wm geometry .att \
	"+[expr $Xpos+$object_distance*$curzoomvalue/100/2]+$Ypos"

    frame .att.lag
    label .att.lag.label -text "Iteration Lag" -width 15 -anchor w
    entry .att.lag.field -textvar link_lag($link_id) -width 10 \
	-relief sunken
    frame .att.actions
    button .att.actions.ok -text "Set" \
	-command "destroy .att"
    button .att.actions.cancel -text "Cancel" \
	-command "destroy .att\n set link_lag($link_id) $link_lag($link_id)\n Display_Timing $win $link_id"
    pack .att.lag.label .att.lag.field -side left -fill x
    pack .att.actions.ok .att.actions.cancel -side left -fill x
    pack .att.lag .att.actions -side top

    bind .att <Return> {.att.actions.ok invoke}
    bind .att <Escape> {.att.actions.cancel invoke}

#    grab .att

}

######################################################################
# Edit_Task_Attributes -
# Create a user dialog frame for specifying the attributes of a task.
# The attributes are: its name, its execution time, 
#
# Parameters:
# - win		the canvas containing the object
# - task_id	the id of the object
# - Xpos	the x coordinate of the object, used to position the popup
# - Ypos	the y coordinate of the object
#
proc Edit_Task_Attributes { win task_id Xpos Ypos } {
    global LawParams TaskTime name curzoomvalue object_distance

    if [catch {toplevel .att}] { return }

    wm title .att "Attributes of task $task_id"

    wm geometry .att \
	"+[expr $Xpos+$object_distance*$curzoomvalue/100/2]+$Ypos"

    frame .att.name
    label .att.name.label -text "Name" -width 15 -anchor w
    entry .att.name.field -textvar name($task_id) -width 20 \
	-relief sunken

    frame .att.time
    label .att.time.label -text "Execution Time" -width 15 -anchor w

    # create law browsing area, setting the parameters to that of 
    # the link
    array set LawParams $TaskTime($task_id)
    set lb [Law_Browser .att.time LawParams]

    frame .att.actions
    button .att.actions.ok -text "Set" \
	-command "Display_Name $win $task_id\n SetTaskLaw $task_id\n Display_Timing $win $task_id\n destroy .att"
    button .att.actions.cancel -text "Cancel" \
	-command "destroy .att\n set name($task_id) {$name($task_id)}"

    pack .att.name.label .att.name.field -side left 
    pack .att.time.label -side left -pady 3 -anchor n
    pack $lb -side left
    pack .att.actions.ok .att.actions.cancel -side left
    pack .att.name .att.time -side top -fill x
    pack .att.actions -side top

    bind .att <Return> {.att.actions.ok invoke}
    bind .att <Escape> {.att.actions.cancel invoke}

#    grab .att

}

######################################################################
# Edit_Object_Attributes -
# Open the window for editing some given object's attributes.
# Calls the proper type-specific editing procedure. 
#
# Parameters:
# - win		the canvas containing the object
# - obj_id	the id of the object
# - Xpos	the x coordinate of the object, used to position the popup
# - Ypos	the y coordinate of the object
#
proc Edit_Object_Attributes { win obj_id Xpos Ypos } {
    global DEBUG

    if $DEBUG {
	puts "Editing Attributes of $obj_id"
    }

    if [Is_Type $obj_id TASK] {
	Edit_Task_Attributes $win $obj_id $Xpos $Ypos
    } elseif [Is_Type $obj_id LINK] {
	if [Is_Type $obj_id PROG] {
	    Edit_ProgLink_Attributes $win $obj_id $Xpos $Ypos
	} elseif [Is_Type $obj_id DEP] {
	    Edit_DepLink_Attributes  $win $obj_id $Xpos $Ypos
	}
    }
}

######################################################################
# Edit_Current_Object_Attributes -
# Open the window for editing the **current** object's attributes.
# Coordinates (Xpos,Ypos) are passed to the procedure. They are NOT 
# used to find the object, but to position the window close to the object.
#
# Parameters:
# - win		the window where this is happening
# - Xpos	a x coordinate
# - Ypos	a y coorfinate
#
proc Edit_Current_Object_Attributes { win Xpos Ypos } {

    set cur_id [$win find withtag current]

    if { $cur_id != "" } { 
	Edit_Object_Attributes $win $cur_id $Xpos $Ypos
    }
}

######################################################################
# SetTaskLaw -
# Procedure to put into the TaskTime array the parameters of the current law.
# Only the fields relevant to this law are extracted from the richer
# LawParams array.
# Note that the value put in TaskTime(...) is a list because it does
# not work with arrays.
# I wish I could use SetLaw (in Common) instead, but I can't make it work,
# probably for the same reason.
#
# Parameters:
# - index	the index of the array at which to put the law array
#
proc SetTaskLaw { index } {
    global LawDesc LawParams TaskTime

    if [info exists TaskTime($index)] {
 	unset TaskTime($index)
    }
    array set TheLaw "$LawDesc($LawParams(name))"
    foreach field [array names TheLaw] {
	set l($field) $LawParams($field)
    }
    set TaskTime($index) "[array get l]"
    # puts "TaskTime($index) set to $TaskTime($index)"
}

######################################################################
# SetLinkLaw -
# Procedure to put into the LinkTime array the parameters of the current law.
# Only the fields relevant to this law are extracted from the richer
# LawParams array.
# I wish I could use SetLaw (in Common) instead...
#
# Parameters:
# - arr		the array to instanciate
# - index	the index of the array at which to put the law array
#
proc SetLinkLaw { index } {
    global LawDesc LawParams LinkTime

    if [info exists LinkTime($index)] {
	unset LinkTime($index)
    }
    array set TheLaw "$LawDesc($LawParams(name))"
    foreach field [array names TheLaw] {
	set l($field) $LawParams($field)
    }
    set LinkTime($index) [array get l]
    puts "LinkTime($index) set to [array get LinkTime($index)]"
}
