#
# Qualitative analysis module for tged
#
# Module grouping the procedures for "qualitative" analysis
# - checking topological properties
# - liveness
#
######################################################################
# Check_Property -
# General handling procedure for calling a boolean function and displaying
# a message according to the result.
#
# Parameters:
# - test_function	the boolean function to call
# - title	title of the popup window
# - description	a string describing the property, used in the message.
#
proc Check_Property { test_function title description } {

    set res [$test_function]
    if { $res == 1 } {
	set message "This is $description."
    } elseif { $res == "unknown" } {
	set message "Can't tell if it is $description."
    } elseif { $res == 0 } {
	set message "This is not $description."
    }

    if { $res == 0 } {
	switch [tk_dialog .graph.property "Bad Properties" "$message\nDo you wish to see the bad objects?" {} 0 {Yes} {No}] {
	    0 { Display_Bad_Objects .graph.canvas }
	    1 { return }

	}
    } else {
	ReportInfo $title $message
    }
}

#######################################################################
# Check_Liveness -
# Check_Deterministic -
# Front-end procedures to the corresponding boolean procedures
#
# Parameters: none
#
proc Check_Liveness { } {

    Check_Property Is_Live "Liveness Analysis" "an acyclic graph"

}

proc Check_Deterministic { } {

    Check_Property Is_Deterministic "Deterministic Analysis" \
	"a deterministic system"

}

######################################################################
# Is_Live - 
# Boolean function returning true if this is a "live" task graph, that is, 
# an acyclic one.
#
# Parameters: none
# 
proc Is_Live { } {
    global dep_link_list bad_obj_list
    global DEBUG

    set res 1
    set the_nodes {}
    set the_links {}
    # initialize lists and tables
    foreach link $dep_link_list {
	set orig [lindex $link 1]
	set dest [lindex $link 2]	
	if { -1 == [lsearch -exact $the_nodes $orig] } {
	    lappend the_nodes $orig
	}
	if { -1 == [lsearch -exact $the_nodes $dest] } {
	    lappend the_nodes $dest
	}
	set indeg($orig) 0
	set indeg($dest) 0
	set odeg($orig) 0
	set odeg($dest) 0
    }
    foreach link $dep_link_list {	
	set orig [lindex $link 1]
	set dest [lindex $link 2]
	incr odeg($orig) 1
	incr indeg($dest) 1
	lappend the_links $link
    }
    set rem_list {}
    set left_list {}
    foreach node $the_nodes {
	if [expr ( $indeg($node) == 0 ) || ( $odeg($node) == 0 )] {
	    lappend rem_list $node
	} else {
	    lappend left_list $node
	}
    }
    puts "nodes: $the_nodes"
    puts "remove: $rem_list"
    puts "left: $left_list"
    puts "links: $the_links"

    while { [llength $rem_list] > 0 } {
	# screen links 
	set new_links {}
	foreach link $the_links {
	    set orig [lindex $link 1]
	    set dest [lindex $link 2]
	    if [expr ( -1 == [lsearch -exact $rem_list $orig] ) \
		&& ( -1 == [lsearch -exact $rem_list $dest] ) ] {
		lappend new_links $link
	    }
	}
	# recompute degrees
	set the_links $new_links
	foreach node $left_list {
	    set indeg($node) 0
	    set odeg($node) 0
	}
	foreach link $the_links {	
	    set orig [lindex $link 1]
	    set dest [lindex $link 2]
	    incr odeg($orig) 1
	    incr indeg($dest) 1
	}
	# recompute initial/final nodes
	set the_nodes $left_list
	set rem_list {}
	set left_list {}
	foreach node $the_nodes {
	    if [expr ( $indeg($node) == 0 ) || ( $odeg($node) == 0 )] {
		if $DEBUG {
		    puts "Removing $node (i=$indeg($node),o=$odeg($node))"
		}
		lappend rem_list $node
	    } else {
		lappend left_list $node
	    }
	}
	if $DEBUG {
	    puts "remove: $rem_list"
	    puts "left: $left_list"
	    puts "links: $the_links"
	}
    }

    if [expr [llength $left_list] == 0] {
	return 1
    } else {
	set bad_obj_list $left_list
	return 0
    }
}

######################################################################
# Is_Deterministic -
# A boolean function answering true if the current task graph is
# deterministic, that is, has all timing laws constant.
#
# Parameters: none
#
proc Is_Deterministic { } {
    global LinkTime TaskTime task_list dep_link_list bad_obj_list
    global DEBUG

    set ok 1
    set bad_obj_list {}

    foreach task $task_list {
	set task_id [lindex $task 0]
	array set Law $TaskTime($task_id)
	if { $Law(name) != "constant" } {
	    set ok 0
	    lappend bad_obj_list $task_id
	}
    }

    foreach link $dep_link_list {
	set link_id [lindex $link 0]
	array set Law $LinkTime($link_id)
	if { $Law(name) != "constant" } {
	    set ok 0
	    lappend bad_obj_list $link_id
	}
    }

    if $DEBUG {
	puts "DET check: bad objects: \[ $bad_obj_list \]"
    }

    return $ok
}

######################################################################
# Is_Type -
# A boolean function answering true if the object passed is of the type
# passed. This is recognized by checking the presence of the tag "type"
# among the objects'.
#
# Parameters:
# - obj		the object to test
# - type	the type to check
#
proc Is_Type { obj type } {

    set objtags  [.graph.canvas gettags $obj]
    return [expr  -1 != [lsearch -exact $objtags $type]]
}

######################################################################
# Is_Empty_Graph -
# A boolean function answering true if the current Task Graph has no
# objects at all.
#
# Parameters: none
#
proc Is_Empty_Graph { } {
    global task_list

    return [expr [llength $task_list] == 0 ]
}

######################################################################
# Has_Program_Links -
# A boolean function answering true if the current Task Graph has some
# program dependency links.
#
# Parameters: none
#
proc Has_Program_Links { } {
    global prog_link_list

    return [expr [llength $prog_link_list] != 0 ]
}

######################################################################
# Has_Zero_Time -
# A boolean function answering true if the object passed has a timing
# with constant zero value.
#
# Parameters:
# - obj		the object to test
# - type	the type to check
#
proc Has_Zero_Time { obj } {
    global TaskTime LinkTime

    if [Is_Type $obj TASK] {
	array set Law $TaskTime($obj)
    } elseif [Is_Type $obj DEP] {
	array set Law $LinkTime($obj)
    } else { 
	return 0
    }

    return [Is_Zero_Time Law]
}

